/*
 * parmes.h
 *
 *  Created on: 2010/09/21
 *      Author: shu
 */

#ifndef ALIGNER_PARAMETERS_H_
#define ALIGNER_PARAMETERS_H_

#include "common_parameters.h"
#include "sequence_type.h"
#include "score_matrix.h"
#include "score_matrix_reader.h"
#include "translator.h"
#include "encoder.h"
#include "statistics.h"
#include <iostream>
#include <stdint.h>
#include <string>

using namespace std;

class Encoder;

class AlignerParameters : public CommonParameters{
public:
  static const uint32_t kMaxQueryLength = 64;
  static const uint32_t kMaxNumberQueries = (1 << 27);
  static const uint32_t kMaxDbLength = ((1 << 30) - 1);

  AlignerParameters();

  virtual ~AlignerParameters();

  void SetParameters(int argc, char* argv[]);

  SequenceTypePtr GetAlignedType() const {
    return aligned_type_;
  }

  string GetDbFileName() const {
    return db_filename_;
  }

  uint32_t GetMaxNumberHits() const {
    return max_number_hits_;
  }

  uint32_t GetMaxQueriesLength() const {
    return max_queries_length_;
  }

  int GetSearchCutoff() const {
    return search_cutoff_;
  }

  string GetQueriesFileNmae() const {
    return queries_filename_;
  }

  SequenceTypePtr GetQueryType() const {
    return query_type_;
  }

  ScoreMatrix GetScoreMatrix() const {
    return score_matrix_;
  }

  int GetNumberThread() const{
    return number_thread_;
  }

  Translator GetTranslator() const {
    return translator_;
  }

  bool Translated() {
    return typeid(*query_type_) != typeid(*aligned_type_);
  }

  uint32_t GetSearchLength() const {
    return search_length_;
  }

  int GetThreshold() const {
    return threshold_;
  }

  uint32_t GetMaxNumberAlignments() const {
    return max_number_alignments_;
  }

  uint32_t GetMaxNumberOneSubjectAlignments() const {
    return max_number_one_subject_alignments_;
  }

  int GetMaxGappedSize() const {
    return max_gapped_size_;
  }

  int GetOpenGap() const {
    return open_gap_;
  }

  string GetOutputFileName() const {
    return output_filename_;
  }

  int GetExtendGap() const {
    return extend_gap_;
  }

  KarlinParameters GetUnGappedIdealKarlinParameters() const {
    return ungapped_ideal_karlin_parameters_;
  }

  KarlinParameters GetGappedKarlinParameters() const {
    return gapped_karlin_parameters_;
  }

  double GetAlpha() const {
    return alpha_;
  }

  double GetBeta() const {
    return beta_;
  }

  float GetUngappedExtendNormalizedCutoff() const {
    return ungapped_extend_normalized_cutoff_;
  }

  float  GetGappedExtendNormalizedCutoff() const {
    return gapped_extend_normalized_cutoff_;
  }

  float  GetFinalGappedExtendNormalizedCutoff() const {
    return final_gapped_extend_normalized_cutoff_;
  }

  float GetGappedExtendNormalizedTrigger() const {
     return gapped_extend_normalized_trigger_;
  }

  int GetUngappedIdealExtendNominalCutoff() const {
    return ungapped_ideal_extend_nominal_cutoff_;
  }

  int GetGappedIdealExtendNominalCutoff() const {
    return gapped_ideal_extend_nominal_cutoff_;
  }

  int GetFinalGappedIdealExtendNominalCutoff() const {
    return final_gapped_ideal_extend_nominal_cutoff_;
  }

  int GetGappedIdealExtendNominalTrigger() const {
    return gapped_ideal_extend_nominal_trigger_;
  }

  uint32_t GetkNextQueryIdIncrement() const {
    return kNextQueryIdIncrement_;
  }

  uint32_t GetkNextAlignmentIdIncrement() const {
    return kNextAlignmentIdIncrement_;
  }

  double GetEntropyThresholdForReordering() const {
    return entropy_threshold_for_reordering_;
  }

  double GetEntropyThresholdForFiltering() const {
    return entropy_threshold_for_filtering_;
  }

  int GetDeviceId() const {
    return device_id_;
  }

private:
  void SetScoringParameters(SequenceType &type, string name, int open_gap, int extend_gap);

  int number_thread_;
  string queries_filename_;
  string db_filename_;
  string output_filename_;
  SequenceTypePtr query_type_;
  SequenceTypePtr aligned_type_;
  uint32_t search_length_;
  int threshold_;
  int search_cutoff_;
  uint32_t max_number_hits_;
  uint32_t max_queries_length_;
  uint32_t max_number_alignments_;
  uint32_t max_number_one_subject_alignments_;
  int max_gapped_size_;
  int open_gap_;
  int extend_gap_;
  ScoreMatrix score_matrix_;
  Translator translator_;
  KarlinParameters ungapped_ideal_karlin_parameters_;
  KarlinParameters gapped_karlin_parameters_;
  double alpha_;
  double beta_;
  float ungapped_extend_normalized_cutoff_;
  float gapped_extend_normalized_cutoff_;
  float final_gapped_extend_normalized_cutoff_;
  float gapped_extend_normalized_trigger_;
  int ungapped_ideal_extend_nominal_cutoff_;
  int gapped_ideal_extend_nominal_cutoff_;
  int final_gapped_ideal_extend_nominal_cutoff_;
  int gapped_ideal_extend_nominal_trigger_;
  uint32_t kNextQueryIdIncrement_;
  uint32_t kNextAlignmentIdIncrement_;
  double entropy_threshold_for_reordering_;
  double entropy_threshold_for_filtering_;

  int device_id_;
};

#endif /* ALIGNER_PARAMETERS_H_ */
