/*
 * chunk.cpp
 *
 *  Created on: 2010/09/29
 *      Author: shu
 */

#include "sa_sequence_index.h"
#include "fmi_sequence_index.h"
#include "encoder.h"
#include <fstream>
#include <stdint.h>
#include <string>
#include <limits.h>
#include "chunk.h"

using namespace std;

Chunk::Chunk(string filename_prefix) :
  filename_prefix_(filename_prefix), number_sequences_(0), concatenated_sequences_length_(0), names_(NULL), offsets_(NULL),
  concatenated_sequence_(NULL),index_(NULL)
{
  LoadInfomation(filename_prefix_);
}

Chunk::~Chunk() {
  if (names_) {
    delete [] names_;
  }

  if (offsets_) {
    delete [] offsets_;
  }

  if (index_) {
    delete index_;
  }

  if (concatenated_sequence_ != NULL) {
    delete [] concatenated_sequence_;
  }

}

uint32_t Chunk::GetId(uint32_t position) {
#pragma omp critical(load_lock_)
    {
  if (offsets_ == NULL) {
    LoadOffsets(filename_prefix_);
  }
    }
  if (offsets_[number_sequences_ - 1]  <= position && position < concatenated_sequences_length_) {
    return number_sequences_ - 1;
  }

  uint32_t left = 0;
  uint32_t right = number_sequences_ - 2;
  uint32_t mid;
  while (left <= right) {
    mid = (left + right)/2;
    if (offsets_[mid]  <= position && position < offsets_[mid + 1]) {
      return mid;
    } else if (offsets_[mid] < position) {
      left = mid + 1;
    } else {
      right = mid - 1;
    }
  }
  return UINT_MAX;
}

void Chunk::LoadInfomation(string filename_prefix) {
  string filename = GetInformationFileName(filename_prefix);
  ifstream in;
  in.open(filename.c_str(), ios::binary);
  if (in) {
    in.read((char *) &number_sequences_, sizeof(uint32_t));
    in.read((char *) &concatenated_sequences_length_, sizeof(uint32_t));
    in.close();
  }
}

void Chunk::LoadOffsets(string filename_prefix) {
  if (offsets_ != NULL) {
    return;
  }
  string filename = GetOffsetsFileName(filename_prefix);
  ifstream in;
  in.open(filename.c_str(), ios::binary);
  if (in) {
    offsets_ = new uint32_t[number_sequences_];
    in.read((char *) offsets_, sizeof(uint32_t) * number_sequences_);
    in.close();
  }
}

void Chunk::LoadNames(string filename_prefix) {
  ifstream in;
  string line;
  string filename = GetNamesFileName(filename_prefix);
  in.open(filename.c_str());
  if (in) {
    names_ = new string[number_sequences_];
    uint32_t i;
    for (i = 0; i < number_sequences_ && !in.eof(); ++i) {
      std::getline(in, line);
      names_[i] = line;
    }
    if (i < number_sequences_) {
      cerr << "warning : couldn't read all sequence names" << endl;
    }
    in.close();
  }
}

void Chunk::LoadConcatenatedSequence(string filename_prefix) {
  string filename = GetSequencesFileName(filename_prefix);
  ifstream in;
  in.open(filename.c_str(), ios::binary);
  if (in) {
    concatenated_sequence_ = new Code[concatenated_sequences_length_];
    in.read((char *) concatenated_sequence_, sizeof(Code) * concatenated_sequences_length_);
    in.close();
  }
}
void Chunk::LoadIndex(string filename_prefix) {
  string filename = GetIndexFileName(filename_prefix);
  ifstream in;
  in.open(filename.c_str(), ios::binary);
  if (in) {
    index_ = new SaSequenceIndex(in);
    //index_ = new FmiSequenceIndex(in);
  }
}

