/*
 * extender.cpp
 *
 *  Created on: 2010/09/28
 *      Author: shu
 */

#include <iostream>
#include <stdio.h>
#include <vector>
#include <stdint.h>
#include <stdio.h>
#include <algorithm>
#include "aligner.h"
#include "queries.h"
#include "db.h"
#include "aligner_parameters.h"
#include "encoder.h"
#include "extender.h"

using namespace std;

Extender::Extender(const AlignerParameters &parameters, Queries &queries) :
  dp_mem_length_(queries.GetMaxSequenceLength() + 3), dp_mem_(NULL),
  gapped_extend_nominal_cutoff_(parameters.GetGappedIdealExtendNominalCutoff()),
  final_gapped_extend_nominal_cutoff_(parameters.GetFinalGappedIdealExtendNominalCutoff()),
  open_gap_(parameters.GetOpenGap()),
  extend_gap_(parameters.GetExtendGap()), max_gapped_size_(parameters.GetMaxGappedSize()), 
  score_matrix_(parameters.GetScoreMatrix()), matrix_(score_matrix_.GetMatrix()),
  max_query_length_(queries.GetMaxSequenceLength())
{
  dp_mem_ = new DpCell[dp_mem_length_];
  
  open_gap_ += extend_gap_;
  if (open_gap_ > gapped_extend_nominal_cutoff_) {
    gapped_extend_nominal_cutoff_ = open_gap_;
  }
  if (open_gap_ > final_gapped_extend_nominal_cutoff_) {
    final_gapped_extend_nominal_cutoff_ = open_gap_;
  } 
}

Extender::~Extender() {
  if (dp_mem_) {
    delete[] dp_mem_;
  }
}

void Extender::GappedExtend(Query &query, Db &db, const Coordinate &hit, int *score, Coordinate *start, Coordinate *end, EditBlocks *edit_blocks) {
  int s = 0;
  int query_offset = 0;
  int db_offset = 0;
  Code *queries_seq = query.GetSequence();
  Code *db_seq = db.GetConcatenatedSequence();

#if 0
  fprintf(stderr, "extending\n");
  fprintf(stderr, "%d %d\n", hit.queries_offset, hit.db_offset );
#endif

  if (max_gapped_size_ > 0) {
    if (edit_blocks == NULL) {
      s += RestrictedGappedExtendOneSideScoreOnly(&queries_seq[hit.queries_offset - 1],
          &db_seq[hit.db_offset - 1], &query_offset, &db_offset, true);
      start->queries_offset = hit.queries_offset + query_offset - 1;
      start->db_offset = hit.db_offset + db_offset - 1;

      s += RestrictedGappedExtendOneSideScoreOnly(&queries_seq[hit.queries_offset],
          &db_seq[hit.db_offset], &query_offset, &db_offset, false);
      end->queries_offset = hit.queries_offset + query_offset;
      end->db_offset = hit.db_offset + db_offset;
      *score = s;
    } else {
      EditBlocks e;
      edit_blocks->Clear();
      s += RestrictedGappedExtendOneSide(&queries_seq[hit.queries_offset - 1],
          &db_seq[hit.db_offset - 1], &query_offset, &db_offset, true, &e);
      start->queries_offset = hit.queries_offset + query_offset - 1;
      start->db_offset = hit.db_offset + db_offset - 1;
      edit_blocks->Add(e);

      s += RestrictedGappedExtendOneSide(&queries_seq[hit.queries_offset],
          &db_seq[hit.db_offset], &query_offset, &db_offset, false, &e);
      end->queries_offset = hit.queries_offset + query_offset;
      end->db_offset = hit.db_offset + db_offset;
      edit_blocks->Add(e);
      *score = s;
    }

  } else {

    if (edit_blocks == NULL) {
      s += GappedExtendOneSideScoreOnly(&queries_seq[hit.queries_offset - 1],
          &db_seq[hit.db_offset - 1], &query_offset, &db_offset, true);
      start->queries_offset = hit.queries_offset + query_offset - 1;
      start->db_offset = hit.db_offset + db_offset - 1;

      s += GappedExtendOneSideScoreOnly(&queries_seq[hit.queries_offset],
          &db_seq[hit.db_offset], &query_offset, &db_offset, false);
      end->queries_offset = hit.queries_offset + query_offset;
      end->db_offset = hit.db_offset + db_offset;
      *score = s;
    } else {
      EditBlocks e;
      edit_blocks->Clear();
      s += GappedExtendOneSide(&queries_seq[hit.queries_offset - 1],
          &db_seq[hit.db_offset - 1], &query_offset, &db_offset, true, &e);
      start->queries_offset = hit.queries_offset + query_offset - 1;
      start->db_offset = hit.db_offset + db_offset - 1;
      edit_blocks->Add(e);

      s += GappedExtendOneSide(&queries_seq[hit.queries_offset],
          &db_seq[hit.db_offset], &query_offset, &db_offset, false, &e);
      end->queries_offset = hit.queries_offset + query_offset;
      end->db_offset = hit.db_offset + db_offset;
      edit_blocks->Add(e);
      *score = s;
    }

  }
}

void Extender::ClearGapState()
{
  for (uint32_t i = 0; i < gap_state_struct_.size(); ++i) {
    gap_state_struct_[i].used = 0;
  }
}

uint32_t Extender::GetGapStateIndex(uint32_t length) {
  uint32_t l = length + length/3; // Add on about 30% so the end will get reused.
  uint32_t max_length = kMaxStateArrayLength;
  uint32_t array_length = max(max_length, l);
  if (gap_state_struct_.empty()) {
    GapState s;
    s.used = 0;
    gap_state_struct_.push_back(s);
    gap_state_struct_[0].state_array.resize(array_length);
    return 0;
  } else {
    uint32_t return_i = gap_state_struct_.size();
    for (uint32_t i = 0; i < gap_state_struct_.size(); ++i) {
      if (l < (gap_state_struct_[i].state_array.size() - gap_state_struct_[i].used)) {
        return_i = i;
        break;
      } else if (gap_state_struct_[i].used == 0) {
        gap_state_struct_[i].state_array.resize(array_length);
        return_i = i;
        break;
      }
    }

    if (return_i == gap_state_struct_.size()) {
      GapState s;
      s.used = 0;
      gap_state_struct_.push_back(s);
      return_i = gap_state_struct_.size() - 1;

      gap_state_struct_[return_i].state_array.resize(array_length);
    }
    return return_i;
  }
}

int Extender::GappedExtendOneSide(Code *query, Code *db, int *query_offset, int *db_offset, bool reversed, EditBlocks *edit_blocks) {
  *query_offset = 0;
  *db_offset = 0;
  edit_blocks->Clear();
  if (query[0] == AlignerParameters::kSequenceEnd || db[0] == AlignerParameters::kSequenceEnd) {
    return 0;
  }

  uint32_t number_extra_cells = 0;
  if (extend_gap_ > 0) {
    number_extra_cells = final_gapped_extend_nominal_cutoff_ / extend_gap_ + 3;
  } else {
    number_extra_cells = max_query_length_ + 3;
  }

  if (number_extra_cells > dp_mem_length_) {
   delete [] dp_mem_;
   dp_mem_length_ = number_extra_cells;
   dp_mem_ = new DpCell[dp_mem_length_];
  }

  ClearGapState();
  vector<uint8_t*> edit_script(100);
  vector<uint32_t> edit_script_offset(edit_script.size());

  uint32_t state_index = GetGapStateIndex(number_extra_cells);
  edit_script[0] = &(gap_state_struct_[state_index].state_array[0]);
  edit_script_offset[0] = 0;
  uint8_t *edit_script_array = &(gap_state_struct_[state_index].state_array[0]);


  uint32_t number_letters = score_matrix_.GetNumberLetters();
  int *score_matrix = &(matrix_[0]);
  int query_position = 0;
  int db_position = 0;
  DpCell *score_array = NULL;
  int increment = 0;
  if (reversed) {
    increment = -1;
  } else {
    increment = 1;
  }
  score_array = dp_mem_;

  int array_start = 0;
  int array_end = 0;
  int array_index = 0;
  int array_last_index = 0;

  int score = -open_gap_;
  score_array[0].best = 0;
  score_array[0].best_gap = -open_gap_;
  edit_script_array[0] = kSubstitution;
  query_position = 0;
  for (array_index = 1; query[query_position] != AlignerParameters::kSequenceEnd; ++array_index, query_position
      += increment) {
    if (score < -final_gapped_extend_nominal_cutoff_) {
      break;
    }
    score_array[array_index].best = score;
    score_array[array_index].best_gap = score - extend_gap_;
    score -= extend_gap_;
  }
  array_end = array_index;

  int best_score = 0;
  int prev_score = 0;
  int score_gap_row = 0;
  int score_gap_column = 0;

#if 0
  fprintf(stderr, "\n");
  fprintf(stderr, "      ");
  for (int x = 0; query[x] != AlignerParameters::kSequenceEnd; x += increment) {
    fprintf(stderr, "%3d", query[x]);
  }
  fprintf(stderr, "\n");
#endif

  for (db_position = 0; db[db_position] != AlignerParameters::kSequenceEnd; db_position
      += increment) {
#if 0
    fprintf(stderr, "%3d", db[db_position - increment]);
    for (int x = 0; x < array_start; ++x) {
      fprintf(stderr, "   ");
    }
    for (int x = array_start; x < array_end; ++x) {
      fprintf(stderr, "%3d", score_array[x].best);
      //fprintf(stderr, "%3d", insertion_db_row[x]);
    }
    fprintf(stderr, "\n");
#endif

    if (extend_gap_ > 0) {
      state_index = GetGapStateIndex(array_end - array_start + number_extra_cells);
    } else {
      state_index = GetGapStateIndex(number_extra_cells - array_start);
    }
    uint32_t edit_script_index = db_position*increment;

    if (edit_script_index >= edit_script.size()) {
      uint32_t new_size = edit_script.size()*2;
      edit_script.resize(new_size);
      edit_script_offset.resize(new_size);
    }
    edit_script[edit_script_index] = &(gap_state_struct_[state_index].state_array[gap_state_struct_[state_index].used + 1]);;
    edit_script_offset[edit_script_index] = array_start;
    edit_script_array = edit_script[edit_script_index] - edit_script_offset[edit_script_index];


    score = kInitScore;
    score_gap_row = kInitScore;
    int *matrix_row = score_matrix + db[db_position] * number_letters;
    query_position = array_start * increment;
    prev_score = score_array[array_start].best;
    score_array[array_start].best = score_array[array_start].best_gap;
    score_array[array_start].best_gap -= extend_gap_;
    edit_script_array[array_start] = kGapInQuery + kExtendGapInQuery;
    array_last_index = array_start;

    for (array_index = array_start + 1; array_index < array_end; ++array_index, query_position
        += increment) {
      score = prev_score + matrix_row[query[query_position]];
      prev_score = score_array[array_index].best;
      score_gap_column = score_array[array_index].best_gap;

      uint8_t op = kSubstitution;
      uint8_t op_col = kExtendGapInQuery;
      uint8_t op_row = kExtendGapInDb;

      if (score < score_gap_column) {
        op = kGapInQuery;
        score = score_gap_column;
      }
      if (score < score_gap_row) {
        op = kGapInDb;
        score = score_gap_row;
      }
      if (best_score > score + final_gapped_extend_nominal_cutoff_) {
        if (array_start + 1 == array_index) {
          ++array_start;
        }
        score_array[array_index].best = kInitScore;
        score_array[array_index].best_gap = kInitScore;
        score_gap_row = kInitScore;
      } else {
        array_last_index = array_index;
        if (score > best_score) {
          best_score = score;
          *query_offset = query_position;
          *db_offset = db_position;
        }

        score_gap_column -= extend_gap_;
        score_gap_row -= extend_gap_;
        if (score_gap_column < score - open_gap_) {
          score_array[array_index].best_gap = score - open_gap_;
        } else {
          score_array[array_index].best_gap = score_gap_column;
          op += op_col;
        }

        if (score_gap_row < score - open_gap_) {
          score_gap_row = score - open_gap_;
        } else {
          op += op_row;
        }
        edit_script_array[array_index] = op;
        score_array[array_index].best = score;
      }
    }

    if (array_start + 1 == array_end) {
      break;
    }

    if (array_last_index < array_end - 1) {
      array_end = array_last_index + 1;
    } else {
      while (score_gap_row >= (best_score - final_gapped_extend_nominal_cutoff_)
          && query[query_position] != AlignerParameters::kSequenceEnd) {
        score_array[array_end].best = score_gap_row;
        score_array[array_end].best_gap = score_gap_row - open_gap_;
        edit_script_array[array_end] = kGapInQuery + kExtendGapInQuery;
        score_gap_row -= extend_gap_;
        ++array_end;
        query_position += increment;
      }
      if (query[query_position] != AlignerParameters::kSequenceEnd) {
        score_array[array_end].best = kInitScore;
        score_array[array_end].best_gap = kInitScore;
        ++array_end;
      }
    }

    gap_state_struct_[state_index].used += array_end - edit_script_offset[edit_script_index];
  }

  PickOptimalPath(edit_script, edit_script_offset, *query_offset, *db_offset, increment, edit_blocks);
  return best_score;
}

int Extender::GappedExtendOneSideScoreOnly(Code *query, Code *db, int *query_offset,
    int *db_offset, bool reversed) {
  *query_offset = 0;
  *db_offset = 0;
  if (query[0] == AlignerParameters::kSequenceEnd || db[0] == AlignerParameters::kSequenceEnd) {
    return 0;
  }
  uint32_t number_letters = score_matrix_.GetNumberLetters();
  int *score_matrix = &(matrix_[0]);
  int query_position = 0;
  int db_position = 0;
  DpCell *score_array = NULL;
  int increment = 0;
  if (reversed) {
    increment = -1;
  } else {
    increment = 1;
  }
  score_array = dp_mem_;

  int array_start = 0;
  int array_end = 0;
  int array_index = 0;
  int array_last_index = 0;

  int score = -open_gap_;
  score_array[0].best = 0;
  score_array[0].best_gap = -open_gap_;
  query_position = 0;
  for (array_index = 1; query[query_position] != AlignerParameters::kSequenceEnd; ++array_index, query_position
      += increment) {
    if (score < -gapped_extend_nominal_cutoff_) {
      break;
    }
    score_array[array_index].best = score;
    score_array[array_index].best_gap = score - extend_gap_;
    score -= extend_gap_;
  }
  array_end = array_index;
  int best_score = 0;
  int prev_score = 0;
  int score_gap_row = 0;
  int score_gap_column = 0;

#if 0
  fprintf(stderr, "\n");
  fprintf(stderr, "      ");
  for (int x = 0; query[x] != AlignerParameters::kSequenceEnd; x += increment) {
    fprintf(stderr, "%3d", query[x]);
  }
  fprintf(stderr, "\n");
#endif

  for (db_position = 0; db[db_position] != AlignerParameters::kSequenceEnd; db_position
      += increment) {
#if 0
    fprintf(stderr, "%3d", db[db_position - increment]);
    for (int x = 0; x < array_start; ++x) {
      fprintf(stderr, "   ");
    }
    for (int x = array_start; x < array_end; ++x) {
      fprintf(stderr, "%3d", score_array[x].best);
      //fprintf(stderr, "%3d", insertion_db_row[x]);
    }
    fprintf(stderr, "\n");
#endif

    score = kInitScore;
    score_gap_row = kInitScore;
    int *matrix_row = score_matrix + db[db_position] * number_letters;
    query_position = array_start * increment;
    prev_score = score_array[array_start].best;
    score_array[array_start].best = score_array[array_start].best_gap;
    score_array[array_start].best_gap -= extend_gap_;
    array_last_index = array_start;

    for (array_index = array_start + 1; array_index < array_end; ++array_index, query_position
        += increment) {
      score = prev_score + matrix_row[query[query_position]];
      prev_score = score_array[array_index].best;
      score_gap_column = score_array[array_index].best_gap;

      if (score < score_gap_column) {
        score = score_gap_column;
      }
      if (score < score_gap_row) {
        score = score_gap_row;
      }
      if (best_score > score + gapped_extend_nominal_cutoff_) {
        if (array_start + 1 == array_index) {
          ++array_start;
        }
        score_array[array_index].best = kInitScore;
        score_array[array_index].best_gap = kInitScore;
        score_gap_row = kInitScore;
      } else {
        array_last_index = array_index;
        if (score > best_score) {
          best_score = score;
          *query_offset = query_position;
          *db_offset = db_position;
        }

        score_array[array_index].best_gap = max(score - open_gap_,
            score_gap_column - extend_gap_);
        score_gap_row = max(score - open_gap_, score_gap_row - extend_gap_);
        score_array[array_index].best = score;
      }
    }

    if (array_start + 1 == array_end) {
      break;
    }

    if (array_last_index < array_end - 1) {
      array_end = array_last_index + 1;
    } else {
      while (score_gap_row >= (best_score - gapped_extend_nominal_cutoff_)
          && query[query_position] != AlignerParameters::kSequenceEnd) {
        score_array[array_end].best = score_gap_row;
        score_array[array_end].best_gap = score_gap_row - open_gap_;
        score_gap_row -= extend_gap_;
        ++array_end;
        query_position += increment;
      }
      if (query[query_position] != AlignerParameters::kSequenceEnd) {
        score_array[array_end].best = kInitScore;
        score_array[array_end].best_gap = kInitScore;
        ++array_end;
      }
    }
  }

  return best_score;
}

int Extender::RestrictedGappedExtendOneSide(Code *query, Code *db, int *query_offset, int *db_offset, bool reversed, EditBlocks *edit_blocks) {
  *query_offset = 0;
  *db_offset = 0;
  edit_blocks->Clear();
  if (query[0] == AlignerParameters::kSequenceEnd || db[0] == AlignerParameters::kSequenceEnd) {
    return 0;
  }

  ClearGapState();
  vector<uint8_t*> edit_script(100);
  vector<uint32_t> edit_script_offset(edit_script.size());

  uint32_t state_index = GetGapStateIndex(max_gapped_size_ + 3);
  edit_script[0] = &(gap_state_struct_[state_index].state_array[0]);
  edit_script_offset[0] = 0;
  uint8_t *edit_script_array = &(gap_state_struct_[state_index].state_array[0]);


  uint32_t number_letters = score_matrix_.GetNumberLetters();
  int *score_matrix = &(matrix_[0]);
  int query_position = 0;
  int db_position = 0;
  DpCell *score_array = NULL;
  int increment = 0;
  if (reversed) {
    increment = -1;
  } else {
    increment = 1;
  }
  score_array = dp_mem_;

  int array_start = 0;
  int array_end = 0;
  int array_index = 0;

  int score = -open_gap_;
  score_array[0].best = 0;
  score_array[0].best_gap = -open_gap_;
  edit_script_array[0] = kSubstitution;
  query_position = 0;
  for (array_index = 1; query[query_position] != AlignerParameters::kSequenceEnd && array_index <= max_gapped_size_; ++array_index, query_position
      += increment) {
    if (score < -final_gapped_extend_nominal_cutoff_) {
      break;
    }
    score_array[array_index].best = score;
    score_array[array_index].best_gap = score - extend_gap_;
    score -= extend_gap_;
  }
  array_end = array_index;

  int best_score = 0;
  int array_best_score = 0;
  int prev_score = 0;
  int score_gap_row = 0;
  int score_gap_column = 0;

#if 0
  fprintf(stderr, "\n");
  fprintf(stderr, "      ");
  for (int x = 0; query[x] != AlignerParameters::kSequenceEnd; x += increment) {
    fprintf(stderr, "%3d", query[x]);
  }
  fprintf(stderr, "\n");
#endif

  for (db_position = 0; db[db_position] != AlignerParameters::kSequenceEnd; db_position
      += increment) {
#if 0
    fprintf(stderr, "%3d", db[db_position - increment]);
    for (int x = 0; x < array_start; ++x) {
      fprintf(stderr, "   ");
    }
    for (int x = array_start; x < array_end; ++x) {
      fprintf(stderr, "%3d", score_array[x].best);
      //fprintf(stderr, "%3d", insertion_db_row[x]);
    }
    fprintf(stderr, "\n");
#endif

    if (!(db_position * increment < max_gapped_size_)) {
      ++array_start;
    }

    state_index = GetGapStateIndex(array_end - array_start);
    uint32_t edit_script_index = db_position*increment;

    if (edit_script_index >= edit_script.size()) {
      uint32_t new_size = edit_script.size()*2;
      edit_script.resize(new_size);
      edit_script_offset.resize(new_size);
    }
    edit_script[edit_script_index] = &(gap_state_struct_[state_index].state_array[gap_state_struct_[state_index].used + 1]);
    edit_script_offset[edit_script_index] = array_start;
    edit_script_array = edit_script[edit_script_index] - edit_script_offset[edit_script_index];


    score = kInitScore;
    score_gap_row = kInitScore;
    array_best_score = kInitScore;
    int *matrix_row = score_matrix + db[db_position] * number_letters;

    query_position = array_start * increment;
    prev_score = score_array[array_start].best;
    score_array[array_start].best = score_array[array_start].best_gap;
    score_array[array_start].best_gap -= extend_gap_;
    edit_script_array[array_start] = kGapInQuery + kExtendGapInQuery;

    for (array_index = array_start + 1; array_index < array_end; ++array_index, query_position
        += increment) {
      score = prev_score + matrix_row[query[query_position]];
      prev_score = score_array[array_index].best;
      score_gap_column = score_array[array_index].best_gap;

      uint8_t op = kSubstitution;
      uint8_t op_col = kExtendGapInQuery;
      uint8_t op_row = kExtendGapInDb;

      if (score < score_gap_column) {
        op = kGapInQuery;
        score = score_gap_column;
      }
      if (score < score_gap_row) {
        op = kGapInDb;
        score = score_gap_row;
      }
      if (score > array_best_score) {
        array_best_score = score;
      }

      if (score > best_score) {
        best_score = score;
        *query_offset = query_position;
        *db_offset = db_position;
      }
      
      score_gap_column -= extend_gap_;
      score_gap_row -= extend_gap_;
      if (score_gap_column < score - open_gap_) {
        score_array[array_index].best_gap = score - open_gap_;
      } else {
        score_array[array_index].best_gap = score_gap_column;
        op += op_col;
      }

      if (score_gap_row < score - open_gap_) {
        score_gap_row = score - open_gap_;
      } else {
        op += op_row;
      }
      edit_script_array[array_index] = op;
      score_array[array_index].best = score;
    }

    if (best_score > array_best_score + final_gapped_extend_nominal_cutoff_) {
      break;
    }
    if (query[query_position] != AlignerParameters::kSequenceEnd) {
      score_array[array_end].best = score_gap_row;
      score_array[array_end].best_gap = score_gap_row - open_gap_;
      ++array_end;
    }
    gap_state_struct_[state_index].used += array_end - edit_script_offset[edit_script_index];
  }

  PickOptimalPath(edit_script, edit_script_offset, *query_offset, *db_offset, increment, edit_blocks);
  return best_score;
}

int Extender::RestrictedGappedExtendOneSideScoreOnly(Code *query, Code *db, int *query_offset, int *db_offset, bool reversed) {
  *query_offset = 0;
  *db_offset = 0;
  if (query[0] == AlignerParameters::kSequenceEnd || db[0] == AlignerParameters::kSequenceEnd) {
    return 0;
  }
  uint32_t number_letters = score_matrix_.GetNumberLetters();

  int *score_matrix = &(matrix_[0]);
  int query_position = 0;
  int db_position = 0;
  DpCell *score_array = NULL;
  int increment = 0;
  if (reversed) {
    increment = -1;
  } else {
    increment = 1;
  }
  score_array = dp_mem_;

  int array_start = 0;
  int array_end = 0;
  int array_index = 0;

  int score = -open_gap_;
  score_array[0].best = 0;
  score_array[0].best_gap = -open_gap_;
  query_position = 0;
  for (array_index = 1; query[query_position] != AlignerParameters::kSequenceEnd && array_index <= max_gapped_size_; ++array_index, query_position
      += increment) {
    if (score < -gapped_extend_nominal_cutoff_) {
      break;
    }
    score_array[array_index].best = score;
    score_array[array_index].best_gap = score - extend_gap_;
    score -= extend_gap_;
  }
  array_end = array_index;
  int best_score = 0;
  int array_best_score = 0;
  int prev_score = 0;
  int score_gap_row = 0;
  int score_gap_column = 0;

#if 0
  fprintf(stderr, "\n");
  fprintf(stderr, "      ");
  for (int x = 0; query[x] != AlignerParameters::kSequenceEnd; x += increment) {
    fprintf(stderr, "%3d", query[x]);
  }
  fprintf(stderr, "\n");
#endif

  for (db_position = 0; db[db_position] != AlignerParameters::kSequenceEnd; db_position
      += increment) {
#if 0
    fprintf(stderr, "%3d", db[db_position - increment]);
    for (int x = 0; x < array_start; ++x) {
      fprintf(stderr, "   ");
    }
    for (int x = array_start; x < array_end; ++x) {
      fprintf(stderr, "%3d", score_array[x].best);
      //fprintf(stderr, "%3d", insertion_db_row[x]);
    }
    fprintf(stderr, "\n");
#endif

    if (!(db_position * increment < max_gapped_size_)) {
      ++array_start;
    }

    score = kInitScore;
    score_gap_row = kInitScore;
    array_best_score = kInitScore;
    int *matrix_row = score_matrix + db[db_position] * number_letters;

    query_position = array_start * increment;
    prev_score = score_array[array_start].best;
    score_array[array_start].best = score_array[array_start].best_gap;
    score_array[array_start].best_gap -= extend_gap_;

    for (array_index = array_start + 1; array_index < array_end; ++array_index, query_position
        += increment) {
      score = prev_score + matrix_row[query[query_position]];
      prev_score = score_array[array_index].best;
      score_gap_column = score_array[array_index].best_gap;

      if (score < score_gap_column) {
        score = score_gap_column;
      }
      if (score < score_gap_row) {
        score = score_gap_row;
      }
      if (score > array_best_score) {
        array_best_score = score;
      }
      if (score > best_score) {
        best_score = score;
        *query_offset = query_position;
        *db_offset = db_position;
      }
      score_array[array_index].best_gap = max(score - open_gap_,
          score_gap_column - extend_gap_);
      score_gap_row = max(score - open_gap_, score_gap_row - extend_gap_);
      score_array[array_index].best = score;

    }

    if (best_score > array_best_score + gapped_extend_nominal_cutoff_) {
      break;
    }

    if (query[query_position] != AlignerParameters::kSequenceEnd) {
      score_array[array_end].best = score_gap_row;
      score_array[array_end].best_gap = score_gap_row - open_gap_;
      ++array_end;
    }
  }
  return best_score;
}

void Extender::PickOptimalPath(const vector<uint8_t*> &edit_script, const vector<uint32_t> &edit_script_offset, int query_position, int db_position, int increment, EditBlocks *edit_blocks) {
  uint8_t op = kSubstitution;
  uint8_t next_op = kSubstitution;

  while (db_position*increment >= 0 && query_position*increment >= 0) {
    uint32_t edit_script_index = db_position*increment;
    next_op = edit_script[edit_script_index][(query_position*increment) - edit_script_offset[edit_script_index] + 1];
    switch(op) {
    case kGapInQuery:
      op = next_op & kEditOpMask;
      if (next_op & kExtendGapInQuery) {
        op = kGapInQuery;
      }
      break;

    case kGapInDb:
      op = next_op & kEditOpMask;
      if (next_op & kExtendGapInDb) {
        op = kGapInDb;
      }
      break;

    default:
      op = next_op & kEditOpMask;
      break;
    }

    switch (op) {
    case kGapInQuery:
      edit_blocks->Add(EditBlocks::kDeletion, 1);
      db_position = db_position - increment;
      break;
    case kGapInDb:
      edit_blocks->Add(EditBlocks::kInsertion, 1);
      query_position = query_position - increment;
      break;
    default:
      edit_blocks->Add(EditBlocks::kSubstitution, 1);
      db_position = db_position - increment;
      query_position = query_position - increment;
      break;
    }
  }
  if (db_position*increment >= 0) {
    edit_blocks->Add(EditBlocks::kDeletion, db_position*increment + 1);
  } else if (query_position*increment >= 0) {
    edit_blocks->Add(EditBlocks::kInsertion, query_position*increment + 1);
  }
  if (increment > 0) {
    edit_blocks->Reverse();
  }
}
