/*
 * amino_mapping_fm.cpp
 *
 *  Created on: 2010/08/02
 *      Author: shu
 */

#include <cstdlib>
#include "common_parameters.h"
#include "assert.h"
#include "logger.h"
#include "aligner.h"
#include "aligner_parameters.h"
#include "db_creator.h"
#include "db_creator_parameters.h"

#ifdef GTEST
#include<gtest/gtest.h>
#endif /* GTEST */


using namespace std;

#ifdef GTEST
int main(int argc, char* argv[])
{
	   testing::InitGoogleTest(&argc, argv);
	   return RUN_ALL_TESTS();
}

#else /* GTEST */

void PrintUsage() {
  Logger *logger = Logger::GetInstance();
  logger->Log(string("GHOSTX - homology search tool. version " + CommonParameters::kVersion));
  logger->Log("\n"
              "Command and Options\n"
              "db: convert a FASTA file to GHOSTX format database files\n"
              "\n"
              "  ghostx db [-i dbFastaFile] [-o dbName] [-l chunkSize]\n"
              "\n"
              "  Options:\n"
              "  (Required)\n"
              "    -i STR    Protein sequences in FASTA format for a database\n"
              "    -o STR    The name of database\n"
              "\n"
              "  (Optional)\n"
              "    -l INT    Chunk size of the database (bytes) [1073741824 (=1GB)]\n"
              "\n"
              "\n"
              "aln:  Search homologues of queries from database\n"
              "\n"
              "ghostx aln [-i queries] [-o output] [-d databes] [-v maxNumAliSub]\n"
              "           [-b maxNumAliQue] [-M scoreMatrix] [-G openGap] [-E extendGap]\n"
              "           [-l CandidatesSize] [-m kNextQueryIdIncrement]\n"
              "           [-n kNextAlignmentIdIncrement] [-s lowerCutoff] [-T UpperCutoff]\n"
              "           [-S searchLength] [-q queryType] [-t databaseType] [-a numThreads]\n"
              "           [-L maxNumHits] [-w maxAliLen]\n"
              "\n"
              "  Options:\n"
              "  (Required)\n"
              "    -i STR    Input query name (must be formatted)\n"
              "    -o STR    Output file\n"
              "    -d STR    database name (must be formatted)\n"
              "\n"
              "  (Optional)\n"
              "    -v INT    Maximum number of alignments for each subject [1]\n"
              "    -b INT    Maximum number of the output for a query [1]\n"
              "\n"
              "    -M STR    Score matrix file[BLOSUM62]\n"
              "    -G INT    Open gap penalty [11]\n"
              "    -E INT    Extend gap penalty [1]\n"
              "\n"
              "    -l INT    Maximun size of the candidates (Bytes) [134217728 (=128MB)]\n"
              "    -m INT    Task size for seed search [1]\n"
              "    -n INT    Task size for traceback [1]\n"
              "    -s INT    Lower limit cutoff score for seed search [1]\n"
              "    -T INT    Upper limit cutoff score for seed search [30]\n"
              "    -S INT    Maximum length of alignments in seed search [8]\n"
              "    -q STR    Query sequence type, p (protein) or d (dna) [p]\n"
              "    -t STR    Database sequence type, p (protein) or d (dna) [p]\n"
              "    -r FLOAT  Entropy threshold for reordering query sequences to enhance parallel performance [3.0]\n"
              "    -f FLOAT  Entropy threshold for filtering query sequences [0.0]\n"
              "    -a INT    The number of threads [1]\n"
              "    -L INT    Maximum number of hits [2097152]\n"
              "    -w INT    Maximum alignment length [50]");
}

int main(int argc, char* argv[]) {
  if (argc < 2 ||
      strcmp(argv[1], "-h") == 0 ||
      strcmp(argv[1], "--help") == 0) {
    PrintUsage();
    exit(1);
  }
  printf("[GHOSTX] start ghostx. \n");fflush(stdout);
  if (strcmp(argv[1], "db") == 0) {
    try {
      DbCreatorParameters parameters;
      parameters.SetParameters(argc - 1, argv + 1);
      DbCreator db_creator;
      db_creator.Run(parameters);
    } catch (exception &e) {
      Logger *logger = Logger::GetInstance();
      logger->ErrorLog(e.what());
    }
  } else if (strcmp(argv[1], "aln") == 0) {
    printf("[GHOSTX] ALIGN mode \n");fflush(stdout);
    try {
      AlignerParameters parameters;
      parameters.SetParameters(argc -1, argv + 1);
      if (parameters.GetDeviceId() == AlignerParameters::kCpuId) {
        Aligner aligner;
        aligner.Run(parameters);
      } else {
        assert(!"Can't use GPU");
      }
    } catch (exception &e) {
      Logger *logger = Logger::GetInstance();
      logger->ErrorLog(e.what());
    }
  } else {
    cerr << "[main] unrecognized command " << argv[1] << endl;
    return EXIT_SUCCESS;
  }
  return EXIT_SUCCESS;
}


#endif /* GTEST */
