/*
 *  bit_vector.cpp
 *
 *  Created on: 2009/09/07
 *      Author: shu
 */

#include <iostream>
#include <string.h>
#include <algorithm>
#include "bit_vector.h"
#include "succinct_bit_vector.h"

using namespace std;

const unsigned char SuccinctBitVector::kPopCountArray[] = {
  0,1,1,2,1,2,2,3,1,2,2,3,2,3,3,4,1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  1,2,2,3,2,3,3,4,2,3,3,4,3,4,4,5,2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  2,3,3,4,3,4,4,5,3,4,4,5,4,5,5,6,3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,
  3,4,4,5,4,5,5,6,4,5,5,6,5,6,6,7,4,5,5,6,5,6,6,7,5,6,6,7,6,7,7,8
};

SuccinctBitVector::SuccinctBitVector
(
    BitVector &bit_vector
):
  blocks_(bit_vector.GetBlocks())
{
  uint32_t size = bit_vector.GetSize();
  level_a_size_ = size/kLevelA + 1;
  level_b_size_ = size/kLevelB + 1;

  level_a_ = new uint32_t[level_a_size_];
  level_b_ = new uint8_t[level_b_size_];

  uint32_t rank = 0;
  for (uint32_t i = 0; i <= size; ++i) {
    if (i % kLevelA == 0) {
      level_a_[i/kLevelA] = rank;
    }
    if (i % kLevelB == 0) {
      level_b_[i/kLevelB] = rank - level_a_[i/kLevelA];
    }
    if (i != bit_vector.GetSize() && i % BitVector::kBlockSize == 0) {
      rank += PopCount(blocks_[i/BitVector::kBlockSize]);
    }
  }
}


SuccinctBitVector::SuccinctBitVector
(
    BitVector &bit_vector,
    uint32_t level_a_size,
    uint32_t level_b_size,
    uint32_t level_a[],
    uint8_t level_b[]
):
  blocks_(bit_vector.GetBlocks()),
  level_a_size_(level_a_size),
  level_b_size_(level_b_size),
  level_a_(level_a),
  level_b_(level_b)
{}




SuccinctBitVector::~SuccinctBitVector
(
    // no parameter
)
{
  if (level_a_) {
    delete [] level_a_;
  }

  if (level_b_) {
    delete [] level_b_;
  }
}
