/*
 * translated_query.cpp
 *
 *  Created on: 2011/04/12
 *      Author: shu
 */

#include "query.h"
#include "translator.h"
#include "translated_query.h"
#include "sequence.h"
#include "dna_sequence.h"
#include "protein_sequence.h"
#include "sa_sequence_index.h"
#include "aligner_parameters.h"
#include "statistics.h"
#include "logger.h"
#include <string>
#include <vector>
#include <stdexcept>

#include <iostream>

using namespace std;

TranslatedQuery::TranslatedQuery(const Sequence &sequence, AlignerParameters &parameters) :
  Query(parameters), parameters_(&parameters), encoder_(*(parameters.GetAlignedType())), dna_length_(sequence.GetSequenceData().size()), is_set_statical_parameters_(false)
{
  SetName(sequence);
  SetEncodedSequence(sequence, encoder_, parameters);
  //SetIndex(sequence_, encoder);
  //SetStaticalParameters(sequence_, parameters);
}

void TranslatedQuery::SetName(const Sequence &sequence) {
  name_ = sequence.GetName();
}


void TranslatedQuery::SetEncodedSequence(const Sequence &sequence, const Encoder &encoder, const AlignerParameters &parameters) {
  DnaSequence dna(sequence.GetName(), sequence.GetSequenceData());
  vector<ProteinSequence> proteins;
  parameters.GetTranslator().Translate(dna, proteins);
  sequence_.resize(TranslatedQuery::GetSequenceLength(dna.GetSequenceData().size()));
  sequence_[0] = AlignerParameters::kSequenceEnd;
  uint32_t offset = 1;
  uint32_t base_length = sequence.GetSequenceData().size()/3;
  for (int i = 0; i < 6; ++i) {
    encoder.Encode(proteins[i].GetSequenceData(), &sequence_[offset]);
    offset += proteins[i].GetSequenceData().size();
    if (proteins[i].GetSequenceData().size() < base_length) {
      sequence_[offset] = AlignerParameters::kSequenceEnd;
      ++offset;
    }
    sequence_[offset] = AlignerParameters::kSequenceEnd;
    ++offset;
  }
}

void TranslatedQuery::SetIndex(const vector<Code> &sequence, const Encoder &encoder) {
  index_ = new SaSequenceIndex(&sequence[0], sequence.size(), encoder);
}
void TranslatedQuery::SetStaticalParameters(const vector<Code> &sequence, const AlignerParameters &parameters) {
  Statistics statistics(*(parameters.GetAlignedType()));
  try {
    statistics.CalculateUngappedKarlinParameters(&sequence[0], sequence.size(),
              parameters.GetScoreMatrix(), &ungapped_karlin_parameters_);
  }catch (invalid_argument e) {
    ungapped_karlin_parameters_ = parameters.GetUnGappedIdealKarlinParameters();
  }
  ungapped_extend_nominal_cutoff_ = Statistics::Normalized2Nominal(parameters.GetUngappedExtendNormalizedCutoff(), ungapped_karlin_parameters_);
  if (ungapped_extend_nominal_cutoff_ > parameters.GetUngappedIdealExtendNominalCutoff()) {
    ungapped_extend_nominal_cutoff_ = parameters.GetUngappedIdealExtendNominalCutoff();
  }
  gapped_extend_nominal_cutoff_ = parameters.GetGappedIdealExtendNominalCutoff();
  gapped_extend_nominal_trigger_ = Statistics::Normalized2Nominal(parameters.GetGappedExtendNormalizedTrigger(), ungapped_karlin_parameters_);
  if (gapped_extend_nominal_trigger_ > parameters.GetGappedIdealExtendNominalTrigger()) {
    gapped_extend_nominal_trigger_ = parameters.GetGappedIdealExtendNominalTrigger();
  }
  is_set_statical_parameters_ = true;
}
